#include <time.h>
#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>

typedef struct{
	int nbCars, nbClasses, nbOptions;
	int *nbCarsInClass, **requires, *classOf, *firstValue, *capacity, *frequency;
	int *totReq;
	int **reqOpt;
} carSeqInstance;

void getData(char* file_name, carSeqInstance* c){
	// Read car sequencing instance contained in file_name and initialize c
	FILE* fd;
	int i, j, k, nbReq;
	
	if ((fd=fopen(file_name,"r"))==NULL){printf("ERROR: Cannot open ascii input file %s", file_name); return;};
	fscanf(fd, "%d%d%d", &(c->nbCars), &(c->nbOptions), &(c->nbClasses));
	printf("nbCars=%d, nbOptions=%d, nbClasses=%d\n",c->nbCars,c->nbOptions,c->nbClasses);
	
	c->capacity = (int*)calloc(c->nbOptions,sizeof(int));
	c->frequency = (int*)calloc(c->nbOptions,sizeof(int));
	c->totReq = (int*)calloc(c->nbOptions,sizeof(int));
	c->nbCarsInClass = (int*)calloc(c->nbClasses,sizeof(int));
	c->firstValue = (int*)calloc(c->nbClasses,sizeof(int));
	c->requires = (int**)calloc(c->nbClasses,sizeof(int*));
	c->reqOpt = (int**)calloc(c->nbClasses,sizeof(int*));
	c->classOf = (int*)calloc(c->nbCars,sizeof(int));
	
	for (i=0; i<c->nbOptions; i++) 
		fscanf(fd,"%d",&(c->capacity[i]));
	for (i=0; i<c->nbOptions; i++){
		fscanf(fd,"%d",&(c->frequency[i]));
		c->totReq[i] = 0;
	}
	k = 0;
	for (i=0; i<c->nbClasses; i++) {
		c->requires[i] = (int*)calloc(c->nbOptions,sizeof(int));
		c->reqOpt[i] = (int*)calloc(c->nbOptions+1,sizeof(int));
		fscanf(fd,"%d%d",&j,&(c->nbCarsInClass[i]));
		nbReq = 0;
		for (j=0; j<c->nbOptions; j++){
			fscanf(fd,"%d",&(c->requires[i][j]));
			if (c->requires[i][j] > 0){
				c->reqOpt[i][nbReq] = j;
				nbReq++;
				c->totReq[j] += c->nbCarsInClass[i];
			}
		}
		c->reqOpt[i][nbReq] = -1;
		c->firstValue[i]=k;
		for (j=0; j<c->nbCarsInClass[i]; j++){
			c->classOf[k]=i;
			k++;
		}
	}
	fclose(fd);
}

int usage (char *exec) {
	printf("\nUsage :\n\n\t%s -a (alpha: int) -b (beta: int) -r (rho1: float) -R (rho2: float) -c (max nb cycle: int) -n (nb ants: int) -m (tau1 min: float) -M (tau1 max: float) -t (strategy: 1 if ACO(tau1,utilRate); 2 if ACO(tau1,tau2)) -f (filename: string) -v display-frequency -s (seed: positive int)\n\n",exec);
	return(0);
}

int parse(int* alpha, int* beta, float* rho1, float* rho2, int* maxCycles, int* nbAnts, float* tau1Min, float* tau1Max, int* verbose, int* displayFreq, char* fileName, int* seed, int* strategy, char* argv[], int argc){
// read arguments to initialize parameters
	char ch;
	extern char* optarg;
	while ( (ch = getopt(argc, argv, "a:b:r:R:c:n:m:M:v:f:?:h:s:t:"))!=-1 ) {
		switch(ch) {
			case 'a': *alpha=atoi(optarg); break;
			case 'b': *beta=atoi(optarg); break;
			case 'r': *rho1=atof(optarg); break;
			case 'R': *rho2=atof(optarg); break;
			case 's': *seed=atoi(optarg); break;
			case 't': *strategy=atoi(optarg); break;
			case 'c': *maxCycles=atoi(optarg); break;
			case 'n': *nbAnts=atoi(optarg); break;
			case 'm': *tau1Min=atof(optarg); break;
			case 'M': *tau1Max=atof(optarg); break;
			case 'v': *verbose=1; *displayFreq=atoi(optarg); break;
			case 'f': strncpy(fileName, optarg, 254); break;
			case '?':
			case 'h':
			default: usage(argv[0]); return(1);
		}
	}
	return(0);
}

void displayCar(int car, carSeqInstance *c){
	int k = c->classOf[car];
	printf("%2d ",k);
}

float myPow(float x, int y){
	// returns x^y (faster than the pow function of math.h...)
	float p;
	if (y==0) return 1;
	if (y==1) return x;
	p = myPow(x,y/2);
	if (y%2==0) return p*p;
	return p*p*x;
}

int choose(float *p, int nbCand, float f){
	/* for i in 0..nbCand-1], p[i] = sum_{j<i} tau1[j]^alpha */
	/* returns k with probability (p[k]-p[k-1])/p[nbCand-1] */
	int left=0;
	int right=nbCand-1;
	int k;
	float total=p[nbCand-1];
	while (left<right){
		k=(left+right+1)/2;
		if (f<p[k-1]/total) right=k-1;
		else if (f>p[k]/total) left=k+1;
		else return k;
	}
	return left;
}

int walkDoublePhero(carSeqInstance* c, float** tau1, float* tau2, int alpha, int beta, float rho2, int* seq){
	// input: c = a car sequencing instance
	//        tau1 = first pheromone structure
	//        tau2 = second pheromone structure
	//        alpha and beta = parameters
	// output: seq = sequence of cars built by an ant w.r.t. ACO(tau1,tau2)
	// returns the number of constraint violations
	float total;
	int nbAssignedCars, j, k, l, cc, min, nbViolations;
	int nbCandidates = 0;
	float p[c->nbClasses];
	int candidates[c->nbClasses], nbSelected[c->nbClasses], cpt[c->nbOptions];
	int first, next[c->nbClasses], nbViolationsClass[c->nbClasses];
	
	nbViolations = 0;
	
	// choice of the first car
	total = 0;
	first = 0;
	for (j=0; j<c->nbClasses; j++){
		nbSelected[j] = 0;
		next[j] = j+1;
		p[j] = myPow(tau2[j],beta) + total;
		total = p[j];
	}
	next[j-1] = -1;
	k = choose(p,j,drand48());
	seq[0] = c->firstValue[k];
	nbSelected[k]++;
	if (c->nbCarsInClass[k]==1){
		if (k==0) first = 1;
		else next[k-1] = next[k];
	}
	for (l=0; l<c->nbOptions; l++) cpt[l] = c->requires[k][l];
	
	// choice of the next cars
	for (nbAssignedCars=1; nbAssignedCars<c->nbCars-1; nbAssignedCars++){
		min = c->nbOptions+1;
		for (j=first; j>=0; j=next[j]){
			nbViolationsClass[j] = 0;
			for (l=0; c->reqOpt[j][l]>=0; l++) 
				if (cpt[c->reqOpt[j][l]]+1-c->capacity[c->reqOpt[j][l]]>0) nbViolationsClass[j]++;
			if (nbViolationsClass[j] < min){
				min=nbViolationsClass[j];
				candidates[0]=c->firstValue[j]+nbSelected[j];
				p[0]=myPow(tau1[seq[nbAssignedCars-1]][candidates[0]],alpha)*myPow(tau2[j],beta);
				total = p[0];
				nbCandidates=1;
			}
			else if (nbViolationsClass[j] == min){
				candidates[nbCandidates] = c->firstValue[j]+nbSelected[j];
				p[nbCandidates]=myPow(tau1[seq[nbAssignedCars-1]][candidates[nbCandidates]],alpha)
				*myPow(tau2[j],beta) + total;
				total = p[nbCandidates];
				nbCandidates++;
			}
		}
		seq[nbAssignedCars] = candidates[choose(p,nbCandidates,drand48())];
		k = c->classOf[seq[nbAssignedCars]];
		if (min>0) // all classes violate constraints => add pheromone on tau2
			for (cc=first; cc>=0; cc=next[cc]) tau2[cc] += nbViolationsClass[cc];
		for (l=0; l<c->nbOptions; l++){
			if (c->requires[k][l]==1){
				cpt[l]++;
				if (nbAssignedCars >= c->frequency[l]-1){
					if (cpt[l] > c->capacity[l]) nbViolations++;
					cpt[l] -= c->requires[c->classOf[seq[nbAssignedCars-c->frequency[l]+1]]][l];
				}
			}
			else if (nbAssignedCars>=c->frequency[l]-1){
				if (cpt[l] > c->capacity[l]) nbViolations++;
				cpt[l] -= c->requires[c->classOf[seq[nbAssignedCars-c->frequency[l]+1]]][l];
			}
		}
		nbSelected[k]++;
		if (nbSelected[k]==c->nbCarsInClass[k]){
			if (first==k) first = next[first];
			else{
				for (j=first; next[j] != k; j=next[j]);
				next[j] = next[k];
			}
		}
	}
	// choice of the last car
	seq[nbAssignedCars] = c->firstValue[first]+nbSelected[first];
	for (l=0; l<c->nbOptions; l++){
		if (c->requires[first][l]==1){
			if (cpt[l]+1 > c->capacity[l]){
				nbViolations++;
				tau2[first]++;
			}
		}
		else if (cpt[l] > c->capacity[l]) nbViolations++;
	}
	/* evaporation of pheromone for the second pheromone structure */
	for (j=0; j<c->nbClasses; j++)
		if (tau2[j]>1) tau2[j] *= 1-rho2;
	return nbViolations;
}


int walkUtilRate(carSeqInstance* c, float** tau, int alpha, int beta, int* seq){
	// input: c = a car sequencing instance
	//        tau = first pheromone structure
	//        alpha and beta = parameters
	// output: seq = sequence of cars built by an ant w.r.t. ACO(tau1,utilRate)
	// returns the number of constraint violations
	float total;
	int nbAssignedCars, i, j, k, l, min, nbViolatedSeq;
	int nbCandidates = 0;
	float p[c->nbClasses];
	int candidates[c->nbClasses], nbSelected[c->nbClasses], cpt[c->nbOptions];
	int first, next[c->nbClasses];
	int currentReq[c->nbOptions];
	float utilRate[c->nbOptions];
	float u;
	
	for (j=0; j<c->nbOptions; j++){
		cpt[j] = 0;
		currentReq[j] = c->totReq[j];
		utilRate[j] = (float)(currentReq[j] * c->frequency[j]) / 
		(float)(c->capacity[j] * c->nbCars);
	}
	first = 0;
	nbViolatedSeq = 0;
	// choice of the first car
	total = 0;
	for (j=0; j<c->nbClasses; j++){
		nbSelected[j] = 0;
		next[j] = j+1;
		u = 0;
		for (i=0; c->reqOpt[j][i]>=0; i++) u += utilRate[c->reqOpt[j][i]];
		p[j] = myPow(u,beta) + total;
		total = p[j];
	}
	next[j-1] = -1;
	k = choose(p,j,drand48());
	seq[0] = c->firstValue[k];
	nbSelected[k]++;
	if (c->nbCarsInClass[k]==1){
		if (k==0) first = 1;
		else next[k-1] = next[k];
	}
	for (l=0; c->reqOpt[k][l]>=0; l++){
		cpt[c->reqOpt[k][l]]++;
		currentReq[c->reqOpt[k][l]]--;
		utilRate[c->reqOpt[k][l]] = (float)(currentReq[c->reqOpt[k][l]] * c->frequency[c->reqOpt[k][l]]) / (c->capacity[c->reqOpt[k][l]] * (c->nbCars - 1));
	}
	// choice of the next cars
	for (nbAssignedCars=1; nbAssignedCars<c->nbCars-1; nbAssignedCars++){
		min = c->nbOptions+1;
		for (j=first; j>=0; j=next[j]){
			k = 0;
			u = 0;
			for (l=0; c->reqOpt[j][l]>=0; l++){
				u += utilRate[c->reqOpt[j][l]];
				if (cpt[c->reqOpt[j][l]]+1-c->capacity[c->reqOpt[j][l]]>0) k++;
			}
			if (k<min){
				min=k;
				candidates[0]=c->firstValue[j]+nbSelected[j];
				p[0]=myPow(tau[seq[nbAssignedCars-1]][candidates[0]],alpha)*myPow(u,beta);
				total = p[0];
				nbCandidates=1;
			}
			else if (k==min){
				candidates[nbCandidates] = c->firstValue[j]+nbSelected[j];
				p[nbCandidates]=myPow(tau[seq[nbAssignedCars-1]][candidates[nbCandidates]],alpha)
				*myPow(u,beta) + total;
				total = p[nbCandidates];
				nbCandidates++;
			}
		}
		seq[nbAssignedCars] = candidates[choose(p,nbCandidates,drand48())];
		k = c->classOf[seq[nbAssignedCars]];
		nbSelected[k]++;
		if (nbSelected[k]==c->nbCarsInClass[k]){
			if (first==k) first = next[first];
			else{
				for (j=first; next[j] != k; j=next[j]);
				next[j] = next[k];
			}
		}
		for (l=0; l<c->nbOptions; l++){
			if (c->requires[k][l]==1){
				currentReq[l]--;
				cpt[l]++;
				if (nbAssignedCars >= c->frequency[l]-1){
					if (cpt[l] > c->capacity[l]) nbViolatedSeq++;
					cpt[l] -= c->requires[c->classOf[seq[nbAssignedCars-c->frequency[l]+1]]][l];
				}
			}
			else if (nbAssignedCars>=c->frequency[l]-1){
				if (cpt[l] > c->capacity[l]) nbViolatedSeq++;
				cpt[l] -= c->requires[c->classOf[seq[nbAssignedCars-c->frequency[l]+1]]][l];
			}
			utilRate[l] = (float)(currentReq[l]*c->frequency[l]) /
			(float)(c->capacity[l]*(c->nbCars - nbAssignedCars - 1));
		}
	}
	// choice of the last car
	seq[nbAssignedCars] = c->firstValue[first]+nbSelected[first];
	for (l=0; l<c->nbOptions; l++){
		if (c->requires[first][l]==1){
			if (cpt[l]+1 > c->capacity[l]){
				nbViolatedSeq++;
			}
		}
		else if (cpt[l] > c->capacity[l]) nbViolatedSeq++;
	}
	
	return nbViolatedSeq;
}


int checkEvalSeq(carSeqInstance* c, int* seq){
	// input: c = a car sequencing instance
	//        seq = a sequence of cars
	// returns the number of constraint violations in seq
	int i, j, k, l, nbViolations;
	int cpt;
	
	nbViolations = 0;
	for (i=0; i<c->nbCars; i++){
		k = c->classOf[seq[i]];
		for (l=0; l<c->nbOptions; l++){
			if (i >= c->frequency[l]-1){
				cpt = 0;
				for (j=0; j<c->frequency[l]; j++) cpt += c->requires[c->classOf[seq[i-j]]][l];
				if (cpt > c->capacity[l]) nbViolations++;
			}
		}
	}
	return nbViolations;
}

int main (int argc, char *argv[]) {
	// Parameters
	char fileName[1024];
	int verbose = 1;
	int displayFreq = 200;
	int alpha = 2;
	int beta = 6;
	float rho1 = 0.01;
	int maxCycles = 5000;
	int nbAnts = 30;
	float tau1Min = 0.01;
	float tau1Max = 4;
	float rho2 = 0.03;
	int seed = 1;
	int strategy = 1;
	
	carSeqInstance c;
	int i, j, nbCycles, total;
	int **seq; // seq[i] = sequence computed by ant i
	int *evalSeq;       // evalSeq[i] = nb of violated constraints in seq[i]
	int *bestSeq;     // best computed sequence
	int evalBestSeq;           // nb of violated constraints in bestSeq
	int bestCycle;             // nb of violated costraints in the best sequence of the cycle
	float** tau1; // first pheromone structure
	float* tau2; // second pheromone structure
	
	float qtyLaid;
	int max;
	clock_t c0=clock();
	
 	if( parse(&alpha, &beta, &rho1, &rho2, &maxCycles, &nbAnts, &tau1Min, &tau1Max, 
			  &verbose, &displayFreq, fileName, &seed, &strategy, argv, argc) == 1) return(1);
	if (verbose==1)
		printf("Parameters: alpha=%d beta=%d rho1=%f rho2=%f tau1Min=%f tau1Max=%f nbCycles=%d nbAnts=%d verbose=%d(%d) input=%s seed=%d strategy=%d\n",
			   alpha, beta, rho1, rho2, tau1Min, tau1Max, maxCycles, nbAnts, verbose, displayFreq, fileName, seed, strategy);
	getData(fileName,&c);
	srand48(seed);
	
	seq = (int**)calloc(nbAnts,sizeof(int*));
	for (i=0; i<nbAnts; i++) seq[i]=(int*)calloc(c.nbCars,sizeof(int));
	bestSeq = (int*)calloc(c.nbCars,sizeof(int));
	evalSeq = (int*)calloc(nbAnts,sizeof(int));
	max = c.nbCars * c.nbOptions;
	
	tau1 = (float**)calloc(c.nbCars,sizeof(float*));
	for (i=0; i<c.nbCars; i++){
		tau1[i] = (float*)calloc(c.nbCars,sizeof(float));
		for(j=0; j<c.nbCars; j++) tau1[i][j]=tau1Max;
	}
	tau2 = (float*)calloc(c.nbClasses,sizeof(float));
	for (i=0; i<c.nbClasses; i++) tau2[i] = 1;

	evalBestSeq = max;
	total = 0;
	for (nbCycles=0; ((nbCycles < maxCycles) && (evalBestSeq>0)); nbCycles++) {
		bestCycle = max;
		/* Each ant computes a path */
		for (i=0; ((i<nbAnts) && (evalBestSeq>0)); i++){
			if (strategy==1)
				evalSeq[i]=walkUtilRate(&c,tau1,alpha,beta,seq[i]);
			else
				evalSeq[i]=walkDoublePhero(&c,tau1,tau2,alpha,beta,rho2,seq[i]);
			total += evalSeq[i];
			if (evalSeq[i]<bestCycle){
				bestCycle = evalSeq[i];
				if (bestCycle<evalBestSeq){
					evalBestSeq = bestCycle;
					if (verbose == 1)
						printf("Cycle %d / Walk %d / %fs: new best = %d\n",
							   nbCycles+1,nbCycles*nbAnts+i+1,(float)(clock()-c0)/CLOCKS_PER_SEC,evalBestSeq);
					for (j=0; j<c.nbCars; j++) bestSeq[j] = seq[i][j];
				}
			}
		}
		if (evalBestSeq==0) break;
		
		/* evaporation of pheromone for the first pheromone structure */
		for (i=0; i<c.nbCars; i++)
			for (j=0; j<c.nbCars; j++)
				if (tau1[i][j]>tau1Min) tau1[i][j] *= 1-rho1;
		
		/* pheromone laying on the first pheromone structure */
		for (i=0; i<nbAnts; i++){
			if (evalSeq[i]==bestCycle){
				qtyLaid = 1.0/(float)(1+evalSeq[i]-evalBestSeq);
				for (j=1; j<c.nbCars; j++){
					if (tau1[seq[i][j-1]][seq[i][j]]<tau1Max) tau1[seq[i][j-1]][seq[i][j]] += qtyLaid;
				}
			}
		}
		
		if (((nbCycles+1)%displayFreq==0) && (verbose==1)){
			printf("Cycle %d / %fs : average eval = %f\n",
				   nbCycles+1,(float)(clock()-c0)/CLOCKS_PER_SEC,(float)(total)/(float)(displayFreq*nbAnts));
			total = 0;
		}
	};
	printf("Number of violations = %d\nsolution = ",checkEvalSeq(&c,bestSeq));
	for (i=0; i<c.nbCars; i++) displayCar(bestSeq[i],&c); printf("\n");
	return(0);
}
